import _ from "lodash";
import { Visitor } from "./Visitor";

/**
 * Base class to represent nodes in pseuCo abstract syntax trees (AST)
 */
export abstract class Node {    
    /**
     * The unique name of the AST node.
     */
    readonly nodeName: string = "Node";

    protected _children: Node[];
    private _parent: Node | null = null;

    /**
     * All of `this`' child nodes in the AST.
     */
    public get childNodes(): ReadonlyArray<Node> {
        return this._children;
    }

    /**
     * The parent node of `this` node or `null`, if it is the root node of the AST.
     */
    public get parent(): Node | null {
        return this._parent;
    }

    /**
     * The root node of the AST that `this` node belongs to. ASTs created by the parser always have a `Program` node as root.
     */
    public get rootNode(): Node {
        if (this._parent) {
            return this._parent.rootNode;
        } else {
            return this;
        }
    }

    /**
     * Possibility for users of this class to store custom information for the node.
     */
    public attributes: Record<symbol, any> = {}

    /**
     * Creates a new AST node.
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param children The child nodes of `this` node in the AST.
     */
    constructor(readonly line: number, readonly column: number, ...children: Node[]) {
        this._children = children;
        children.forEach((child) => {
            child._parent = this;
        });
    }

    /**
     * Checks whether the node is part of a monitor declaration.
     * @returns true, if and only if `this` node is a descendant of `Monitor`.
     */
    insideMonitor(): boolean {
        return this.parent != null && this.parent.insideMonitor();
    }

    /**
     * Checks whether the node is part of a struct declaration.
     * @returns true, if and only if `this` node is a descendant of `Struct`.
     */
    insideStruct(): boolean {
        return this.parent != null && this.parent.insideStruct();
    }

    /**
     * Checks whether the node is part of a procedure or main agent declaration.
     * @returns true, if and only if `this` is or has a parent node that is a procedure declaration or the mainAgent declaration.
     */
    insideProcedure(): boolean {
        return this.parent != null && this.parent.insideProcedure();
    }

    /**
     * Checks whether the node is part of a loop body, loop predicate, etc.
     * @returns true, if and only if `this` node is a descendant of a loop node.
     */
    insideLoop(): boolean {
        return this.parent != null && this.parent.insideLoop();
    }

    /**
     * Counts the number of send (`chan <! val`) and receive (`<? chan`) operations in descendant nodes. If `this` node is a send or receive operation, it is included in the count.
     * @returns The number of send or receive operations in descendant nodes.
     */
    usedSendOrReceiveOperators(): number {
        return this.childNodes.reduce(function(acc: number, child: Node) {
            return acc + child.usedSendOrReceiveOperators();
        }, 0);
    }

    /**
     * Checks whether the node is part of a case statement.
     * @returns true, if and only if `this` node is a descendant of `Case`.
     */
    insideCaseStatement(): boolean {
        return this.parent != null && this.parent.insideCaseStatement();
    }

    /**
     * Checks whether some descendant node calls a procedure.
     * @returns true, if and only if a descendant node is a `ProcedureCall`.
     */
    containsProcedureCall(): boolean {
        return this.childNodes.some(function(child) {
            return child.containsProcedureCall();
        });
    }

    /**
     * Computes a string representation of the node (including all descendant nodes), properly indented using `indent`.
     * @param indent The string used to indent lines (e.g., `"\t"` or `"␣␣"`).
     * @returns A string representation of the node (including all descendant nodes), properly indented.
     */
    abstract indentedString(indent: string): string;

    toString(): string {
        return this.indentedString("");
    }

    /**
     * Implementation of the Visitor design pattern.
     * @param visitor An object implementing the `Visitor` pattern.
     * @param context Any object that a concrete visitor implementation wants to use to remember contextual information.
     * @returns The result provided by the visitor.
     */
    abstract accept<T>(visitor: Visitor<T>, context?: unknown): T;

    /**
     * Default indentation string of two whitespace characters.
     */
    static INDENT = "  ";

}