import { Node, ProcedureDeclaration, ConditionDeclaration, VariableDeclarations } from "./index";
import { Visitor } from "./../Visitor";

export type MonitorDeclaration = ProcedureDeclaration | ConditionDeclaration | VariableDeclarations;

/**
 * AST node representing monitor declarations.
 * @example monitor M {
 *   // this.declarations: array of procedure, condition, and variable declarations
 * }
 */
export class Monitor extends Node {
    readonly nodeName = "Monitor";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param name The name of the monitor (e.g., `M`).
     * @param declarations Procedure, condition, and variable declarations defining the behaviour of the monitor.
     */
    constructor(line: number, column: number, readonly name: string, ...declarations: MonitorDeclaration[]) {
        super(line, column, ...declarations);
    }

    /**
     * Procedure, condition, and variable declarations defining the behaviour of the monitor.
     */
    get declarations(): MonitorDeclaration[] {
        return this.childNodes as MonitorDeclaration[];
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitMonitor(this, context);
    }

    override indentedString(_: string): string {
        const declarationString = this.declarations.map(function(declaration) {
            return declaration.indentedString(Node.INDENT);
        }).join("\n");
        return `monitor ${this.name} {\n${declarationString}\n}`;
    }

    override insideMonitor(): boolean {
        return true;
    }
}