import { Node, TypeNode, ProcedureParameter, StatementBlock } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing the declaration of procedures (global, monitor, and struct procedures).
 * @example int inc(int x) {
 *   return x + 1;
 * }
 */
export class ProcedureDeclaration extends Node {
    readonly nodeName = "ProcedureDeclaration";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param resultTypeNode The node representing the return type of the procedure (e.g., `int`)
     * @param name The name of the procedure (e.g., `"inc"`)
     * @param parameters The parameters of the procedure (e.g., [`int x`])
     * @param body The body of the procedure (e.g., `{ return x + 1; }`)
     */
    constructor(line: number, column: number, resultTypeNode: TypeNode, readonly name: string, parameters: ProcedureParameter[], body: StatementBlock) {
        super(line, column, resultTypeNode, ...parameters, body);
    }

    /**
     * The node representing the return type of the procedure
     * @example int
     */
    get resultTypeNode(): TypeNode {
        return this.childNodes[0] as TypeNode;
    }

    /**
     * The parameters of the procedure
     * @example (int x)
     */
    get parameters(): ProcedureParameter[] {
        return this.childNodes.slice(1, this.childNodes.length-1) as ProcedureParameter[];
    }

    /**
     * The body of the procedure
     * @example {
     *   return x + 1;
     * }
     */
    get body(): StatementBlock {
        return this.childNodes[this.childNodes.length-1] as StatementBlock;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitProcedureDeclaration(this, context);
    }

    override indentedString(indent: string): string {
        const parameterString = this.parameters.map(function(parameter) {
            return parameter.toString();
        }).join(", ");
        return `${indent}${this.resultTypeNode.toString()} ${this.name}(${parameterString}) ${this.body.indentedString(indent)}`;
    }

    override insideProcedure(): boolean {
        return true;
    }


    // Backwards Compatibility

    /**
     * @deprecated
     */
    getResultType(): TypeNode {
        return this.resultTypeNode;
    }

    /**
     * @deprecated
     */
    getBody(): StatementBlock {
        return this.body;
    }

    /**
     * @deprecated
     */
    getArgumentCount(): number {
        return this.childNodes.length - 2;
    }

    /**
     * @deprecated
     */
    getArgumentAtIndex(index: number): ProcedureParameter {
        return this.parameters[index];
    }

}

/**
 * @deprecated
 */
export const PCProcedureDecl = ProcedureDeclaration;