import { Node, ProcedureDeclaration, VariableDeclarations } from "./index";
import { Visitor } from "./../Visitor";

export type StructDeclaration = ProcedureDeclaration | VariableDeclarations;

/**
 * AST node representing struct declarations.
 * @example struct S {
 *   // this.declarations: array of procedure and variable declarations
 * }
 */
export class Struct extends Node {
    readonly nodeName = "Struct";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param name The name of the struct (e.g., `S`).
     * @param declarations Procedure and variable declarations defining the behaviour of the struct.
     */
    constructor(line: number, column: number, readonly name: string, ...declarations: StructDeclaration[]) {
        super(line, column, ...declarations);
    }

    /**
     * Procedure and variable declarations defining the behaviour of the struct.
     */
    get declarations(): StructDeclaration[] {
        return this.childNodes as StructDeclaration[];
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitStruct(this, context);
    }

    override indentedString(_: string): string {
        const declarationString = this.declarations.map(function(declaration) {
            return declaration.indentedString(Node.INDENT);
        }).join("\n");
        return `struct ${this.name} {\n${declarationString}\n}`;
    }

    override insideStruct(): boolean {
        return true;
    }
}