import { Node, TypeNode, VariableDeclarator, Program, Monitor, Struct, StatementBlock, MainAgent } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing the declaration of variables.
 * @example int x = 4, y, z = 2;
 */
export class VariableDeclarations extends Node {
    readonly nodeName = "VariableDeclarations";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param typeNode The node representing the type of all declared variables (e.g., `int`)
     * @param declerators The declerators declaring each individual variable (e.g., [`x = 4`, `y`, `z = 2`])
     */
    constructor(line: number, column: number, typeNode: TypeNode, ...declerators:VariableDeclarator[]) {
        super(line, column, typeNode, ...declerators);
    }

    /**
     * The node representing the type of all declared variables.
     * @example int
     */
    get typeNode(): TypeNode {
        return this.childNodes[0] as TypeNode;
    }

    /**
     * The declerators declaring each individual variable
     * @example x = 4, y, z = 2
     */
    get declarators(): VariableDeclarator[] {
        return this.childNodes.slice(1, this.childNodes.length) as VariableDeclarator[];
    }

    /**
     * Variable declarations are considered to be statements if and only if they appear on the top level of a program, a struct declaration, monitor declaration or in a statement block (i.e., at the top-level between pairs of curly braces `{}`). Variable declarations that are not statements appear in the initialisation of a `for`-loop.
     */
    get isStatement(): boolean {
        return this.parent instanceof Program || this.parent instanceof Struct || this.parent instanceof Monitor || this.parent instanceof StatementBlock;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitVariableDeclarations(this, context);
    }

    /**
     * Computes a string representation of the node (including all descendant nodes), properly indented using `indent` and by adding a semicolon if and only if `isStatement` is true.
     * @param indent The string used to indent lines (e.g., `"\t"` or `"␣␣"`).
     * @param isStatement A flag indicating whether the declarations should be considered as a statement.
     * @returns A string representation of the node (including all descendant nodes), properly indented.
     */
    indentedDeclarationString(indent: string, isStatement: boolean): string {
        const declaratorsString = this.declarators.map(function(declarator: VariableDeclarator) {
            return declarator.toString();
        }).join(", ");
        const semicolon = (isStatement) ? ";" : "";
        return `${indent}${this.typeNode.toString()} ${declaratorsString}${semicolon}`;
    }

    override indentedString(indent: string): string {
        return this.indentedDeclarationString(indent, this.isStatement);
    }


    // Backwards compatibility

    /**
     * @deprecated
     */
    getTypeNode(): TypeNode {
        return this.typeNode;
    }

    /**
     * @deprecated
     */
    getDeclarators(): VariableDeclarator[] {
        return this.declarators;
    }

}