import { Node, VariableInitializer, TypeNode, VariableDeclarations } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing, together with `VariableDeclarations`, the declaration of a variable.
 * @example x = 42
 */
export class VariableDeclarator extends Node {
    readonly nodeName;

    /**
     * The identifier of the variable (e.g., `x`)
     */
    readonly identifier: string;

    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param identifier The identifier of the variable (e.g., `x`).
     * @param initializer Optional argument to provide a node defining the initial value of the variable (e.g., `42` or `[1, 5, 3,]` for arrays).
     */
    constructor(line: number, column: number, identifier: string, initializer?: VariableInitializer) {
        if (initializer) {
            super(line, column, initializer);
        } else {
            super(line, column);
        }
        // The initialisation of this class is a bit weird, because TypeScript is quite strict about when `super` must be called.
        this.identifier = identifier;
        this.nodeName = "VariableDeclarator";
    }

    /**
     * A node defining the initial value of the variable (e.g., `42` or `[1, 5, 3,]` for arrays) or *null* for the type's default initial value.
     */
    get initializer(): VariableInitializer | null {
        if (this.childNodes.length > 0) {
            return this.childNodes[0] as VariableInitializer;
        } else {
            return null;
        }
    }

    /**
     * Shortcut to get the AST node representing the type of the variable.
     */
    get typeNode(): TypeNode {
        return (this.parent as VariableDeclarations).typeNode;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitVariableDeclarator(this, context);
    }

    override indentedString(_: string): string {
        let res = this.identifier;
        if (this.initializer != null) {
            res += ` = ${this.initializer.toString()}`;
        }
        return res;
    }



    // Backwards compatibility

    /**
     * @deprecated
     */
    get name(): string {
        return this.identifier;
    }

    /**
     * @deprecated
     */
    getInitializer(): VariableInitializer | null {
        return this.initializer;
    }

    /**
     * @deprecated
     */
    getTypeNode(): TypeNode {
        return this.typeNode;
    }
}