import { Node, Expression, TypeNode, VariableDeclarator } from "./index";
import { Visitor } from "./../Visitor";

export type VariableInitializerChild = Expression | VariableInitializer;

/**
 * AST node representing the initialising value for a variable declaration.
 * @example 42 // for an integer variable
 * [2, 8, 4, ] // for an array variable
 */
export class VariableInitializer extends Node {
    readonly nodeName = "VariableInitializer";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param isUncompletedArray Boolean flag whether the initial value is an incompleted array, i.e., if an array of size *n* is initialised with an array of size *m* ≤ *n*. In such a case, the pseuCo syntax enforces a comma after the last provided value (e.g., `[2, 8, 4, ]` can initialise arrays of size 3 or more). 
     * @param children A list of initialisers for array initialisations or a single expression for all other types.
     */
    constructor(line: number, column: number, readonly isUncompletedArray: boolean, ...children: VariableInitializerChild[]) {
        super(line, column, ...children);
    }

    /**
     * For non-array values this is an expression node representing the initial value. For array types its value is *null*.
     */
    get expression(): Expression | null {
        if (this.childNodes[0] instanceof Expression) {
            return this.childNodes[0] as Expression;
        } else {
            return null;
        }
    }

    /**
     * For array types this is a list of further variable initialisers, each representing an element in the array. For non-array types its value is *null*.
     */
    get arrayInitializers(): VariableInitializer[] | null {
        if (this.expression == null) {
            return this.childNodes as VariableInitializer[];
        } else {
            return null;
        }
    }

    /**
     * Shortcut to get the AST node representing the type of the variable.
     */
    get typeNode(): TypeNode {
        return (this.parent as VariableDeclarator).typeNode;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitVariableInitializer(this, context);
    }

    override indentedString(_: string): string {
        if (this.expression != null) {
            return this.expression.toString();
        } else {
            const comma = (this.isUncompletedArray) ? "," : "";
            const values = this.arrayInitializers!.map(function(initializer) {
                return initializer.toString();
            }).join(", ");
            return `{${values}${comma}}`;
        }
    }


    // Backwards compatibility

    /**
     * @deprecated
     */
    isArray(): boolean {
        return this.expression == null;
    }

    /**
     * @deprecated
     */
    getTypeNode(): TypeNode {
        return this.typeNode;
    }
}