import { Node } from "./index";
import { Visitor } from "./../Visitor";
import { Expression } from "./Expression";

/**
 * AST node representing the left-hand side of an assignment. Can be an identifier or an identifier followed by array indices. The indices may be expression.
 * @example x
 * a[2]
 * a[2][y+1]
 */
export class AssignDestination extends Node {
    readonly nodeName = "AssignDestination";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param identifier The identifier of a variable.
     * @param arrayIndexExpressions Zero or more expression nodes representing indices.
     */
    constructor(line: number, column: number, readonly identifier: string, ...arrayIndexExpressions: Expression[]) {
        super(line, column, ...arrayIndexExpressions);
    }

    /**
     * Zero or more expression nodes representing indices.
     */
    get arrayIndexExpressions(): Expression[] {
        return (this.childNodes as Expression[]);
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitAssignDestination(this, context);
    }

    override indentedString(indent: string): string {
        //const reversedChildren = this.arrayIndexExpressions.slice(0).reverse();
        const indices = this.arrayIndexExpressions.map(function(indexExpression) {
            return `[${indexExpression.toString()}]`;
        }).join("");
        return `${this.identifier}${indices}`;
    }
}