import { Expression, AssignDestination } from "./index";
import { Visitor } from "./../Visitor";

export type AssignOperator = "=" | "+=" | "*=" | "/=" | "-=";

/**
 * AST node representing an assignment.
 * @example x = 42
 * x += 42
 * x -= 42
 * x *= 42
 * x /= 42
 */
export class AssignExpression extends Expression {
    readonly nodeName = "AssignExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param destination The AST node representing the left-hand side of the assignment.
     * @param operator The assignment operator ("=", "+=", "-=", "*=" or "/=")
     * @param valueExpression The AST node representing the right-hand side of the assignment.
     */
    constructor(line: number, column: number, destination: AssignDestination, readonly operator: AssignOperator, valueExpression: Expression) {
        super(line, column, destination, valueExpression);
    }

    /**
     * The AST node representing the left-hand side of the assignment.
     */
    get assignDestination(): AssignDestination {
        return (this.childNodes[0] as AssignDestination);
    }

    /**
     * The AST node representing the right-hand side of the assignment.
     */
    get valueExpression(): Expression {
        return (this.childNodes[1] as Expression);
    }

    override get precedence(): number {
        return 39;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitAssignExpression(this, context);
    }

    override indentedString(indent: string): string {
        return `${this.assignDestination.toString()} ${this.operator} ${this.childToString(this.valueExpression)}`;
    }


    // Backwards Compatibility

    /**
     * @deprecated
     */
    getDestination(): AssignDestination {
        return this.assignDestination;
    }

    /**
     * @deprecated
     */
    getExpression(): Expression {
        return this.valueExpression;
    }
}