import { Expression } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing a conditional expression.
 * @example predicate ? consequence : alternative
 */
export class ConditionalExpression extends Expression {
    readonly nodeName = "ConditionalExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param predicate The AST node representing the boolean expression depending on which to choose the consequence or alternative.
     * @param consequence The AST node representing the expression that is evaluated if the predicate is true.
     * @param alternative The AST node representing the expression that is evaluated if the predicate is false.
     */
    constructor(line: number, column: number, predicate: Expression, consequence: Expression, alternative: Expression) {
        super(line, column, predicate, consequence, alternative);
    }

    /**
     * The AST node representing the boolean expression depending on which to choose the consequence or alternative.
     */
    get predicate(): Expression {
        return this.childNodes[0] as Expression;
    }

    /**
     * The AST node representing the expression that is evaluated if the predicate is true.
     */
    get consequence(): Expression {
        return this.childNodes[1] as Expression;
    }

    /**
     * The AST node representing the expression that is evaluated if the predicate is false.
     */
    get alternative(): Expression {
        return this.childNodes[2] as Expression;
    }

    override get precedence(): number {
        return 45;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitConditionalExpression(this, context);
    }

    override indentedString(indent: string): string {
        return `${this.childToString(this.predicate)} ? ${this.consequence.toString()} : ${this.alternative.toString()}`;
    }
}