import { Node } from "./index";

/**
 * Base class for AST nodes representing expressions.
 */
export abstract class Expression extends Node {
    readonly nodeName: string = "Expression";
    
    /**
     * The operator precedence to disambiguate which operands belong to which operator. For example, an addition has precedence 60 and multiplication 63, thus, multiplication "consumes" operands stronger than addition. The effect is that 3 + 4 * 5 implicitly is evaluated as 3 + (4 * 5) and that `toString` knows about that fact and computes the minimal amount of parentheses for expressions.
     */
    protected abstract get precedence(): number;

    /**
     * Auxiliary method to decide whether the child representation of a child node must be encapsulated in parentheses or not.
     * @param child The child node that shall be converted into a string.
     * @param diff Precedence modification parameter, e.g., to get different behaviour depending on whether `child` is a left or right operand of `this`. The modification parameter is added to `this`' precedence.
     * @returns A string representation of `child`, encapsulated in parentheses if necessary.
     */
    protected childToString(child: Expression, diff: number = 0): string {
        if (this.precedence + diff > child.precedence) {
            return `(${child.toString()})`;
        } else {
            return child.toString();
        }
    }

    // Backwards compatibility

    /**
     * @deprecated
     */
    getPrecedence(): number {
        return this.precedence;
    }
}