import { Expression } from "./index";
import { Visitor } from "./../Visitor";
import { IllegalStateException } from "../../pseuco/constants";

/**
 * The AST node representing literals as expressions.
 * @example 42
 * "Hello"
 * true
 */
export class LiteralExpression extends Expression {
    readonly nodeName = "LiteralExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param valueWrapper The type and value of the literal.
     */
    constructor(line: number, column: number, readonly valueWrapper: ValueWrapper) {
        super(line, column);
    }

    override get precedence(): number {
        return 84;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitLiteralExpression(this, context);
    }

    override indentedString(indent: string): string {
        switch (this.valueWrapper.type) {
            case ValueType.BOOL:
                return (this.valueWrapper.value == true) ? "true" : "false";

            case ValueType.INT:
                return `${this.valueWrapper.value}`;

            case ValueType.STRING:
                return `"${this.valueWrapper.value}"`;
        
            default:
                throw new IllegalStateException("Types other than bool, int and string are not allowed in literal expressions!");
        }
    }


    // Backwards compatibility

    /**
     * @deprecated
     */
    get value(): boolean | number | string {
        return this.valueWrapper.value;
    }
}


/**
 * A wrapper class to capture type and value of literals.
 */
export class ValueWrapper {
    /**
     * 
     * @param type The type of the value.
     * @param value The boolean, number or string value.
     */
    constructor(readonly type: ValueType, readonly value: boolean | number | string) {}

    get isBool(): boolean {
        return this.type == ValueType.BOOL;
    }

    get isInt(): boolean {
        return this.type == ValueType.INT;
    }

    get isString(): boolean {
        return this.type == ValueType.STRING;
    }
}

/**
 * The possible types for values. BOOL, INT or STRING.
 */
export enum ValueType {
    // VOID,
    BOOL,
    INT, 
    STRING, 
    // CHANNEL,
    // ARRAY,
    // MONITOR,
    // STRUCT,
    // LOCK,
    // MUTEX,   // was deprecated, but still exists here to show proper error message to the user
    // CONDITION,
    // PROCEDURE,
    // TYPE,
    // MAINAGENT,
    // AGENT,
    // WILDCARD
}