import { Expression } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing the call of a procedure.
 * @example f()
 */
export class ProcedureCall extends Expression {
    readonly nodeName = "ProcedureCall";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param procedureName The name of the procedure.
     * @param argumentExpressions A list of AST nodes representing the expressions which shall be evaluated as arguments to the procedure.
     */
    constructor(line: number, column: number, readonly procedureName: string, ...argumentExpressions: Expression[]) {
        super(line, column, ...argumentExpressions);
    }

    /**
     * A list of AST nodes representing the expressions which shall be evaluated as arguments to the procedure.
     */
    get argumentExpressions(): Expression[] {
        return this.childNodes as Expression[];
    }

    override get precedence(): number {
        return 75;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitProcedureCall(this, context);
    }

    override indentedString(indent: string): string {
        const argumentsString = this.argumentExpressions.map(function(expression) {
            return expression.toString();
        }).join(", ");
        return `${this.procedureName}(${argumentsString})`;
    }

    override containsProcedureCall(): boolean {
        return true;
    }
}