import { Expression } from "./index";
import { Visitor } from "./../Visitor";

export type RelationalOperator = "<" | ">" | "<=" | ">=";

/**
 * AST node representing the comparison of two expressions.
 * @example x < 0
 * x > 0
 * x <= 0
 * x >= 0
 */
export class RelationalExpression extends Expression {
    readonly nodeName = "RelationalExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param left The AST node representing the left operand.
     * @param operator The operator, i.e., "<", ">", "<=" or ">=".
     * @param right The AST node representing the right operand.
     */
    constructor(line: number, column: number, left: Expression, readonly operator: RelationalOperator, right: Expression) {
        super(line, column, left, right);
    }

    /**
     * The AST node representing the left operand.
     */
    get left(): Expression {
        return this.childNodes[0] as Expression;
    }

    /**
     * The AST node representing the right operand.
     */
    get right(): Expression {
        return this.childNodes[1] as Expression;
    }

    override get precedence(): number {
        return 57;
    }

    override indentedString(indent: string): string {
        return `${this.childToString(this.left)} ${this.operator} ${this.childToString(this.right, 1)}`;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitRelationalExpression(this, context);
    }
}