import { Expression, Node } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing send expressions.
 * @example chan <! val
 */
export class SendExpression extends Expression {
    readonly nodeName = "SendExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param channelExpression The AST node representing the expression that evaluates to the channel to which a value shall be sent.
     * @param valueExpression The AST node representing an expression that evaluates to the value that shall be sent.
     */
    constructor(line: number, column: number, channelExpression: Expression, valueExpression: Expression) {
        super(line, column, channelExpression, valueExpression);
    }

    /**
     * The AST node representing the expression that evaluates to the channel to which a value shall be sent.
     */
    get channelExpression(): Expression {
        return (this.childNodes[0] as Expression);
    }

    /**
     * The AST node representing an expression that evaluates to the value that shall be sent.
     */
    get valueExpression(): Expression {
        return (this.childNodes[1] as Expression);
    }

    override get precedence(): number {
        return 39;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitSendExpression(this, context);
    }

    override indentedString(indent: string): string {
        const channelString = this.childToString(this.channelExpression, 1);
        const valueString = this.childToString(this.valueExpression);
        return `${channelString} <! ${valueString}`;
    }

    override usedSendOrReceiveOperators(): number {
        return this.childNodes.reduce(function(acc: number, child: Node) {
            return acc + child.usedSendOrReceiveOperators();
        }, 1);
    }
}