import { Node, StatementWrapper, StatementExpression, SendExpression, ReceiveExpression, AssignExpression } from "./index";
import { Visitor } from "./../Visitor";

export type CaseCondition = SendExpression | ReceiveExpression | AssignExpression;

/**
 * AST node representing a case inside a select statement.
 * @example case <? chan: { }
 * case x = <? chan: { }
 * case chan <! val: { }
 * default: { }
 */
export class Case extends Node {
    readonly nodeName;
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param statement The body of the case, which is executed if the case can be and is chosen.
     * @param condition The send or receive operation that must not deadlock to choose the case.
     */
    constructor(line: number, column: number, statement: StatementWrapper, condition?: CaseCondition) {
        if (condition == undefined) {
            super(line, column, statement);
        } else {
            super(line, column, statement, condition);
        }
        this.nodeName = "Case";
    }

    /**
     * The body of the case, which is executed if the case can be and is chosen.
     */
    get statement(): StatementWrapper {
        return this.childNodes[0] as StatementWrapper;
    }

    /**
     * The send or receive operation that must not deadlock to choose the case. If the condition is *null*, then this is a default case that can always be chosen.
     */
    get condition(): CaseCondition | null {
        if (this.childNodes.length > 1) {
            return this.childNodes[1] as CaseCondition;
        } else {
            return null;
        }
    }

    /**
     * true if and only if this case is a default case that can always be chosen.
     */
    get isDefaultCase(): boolean {
        return this.childNodes.length == 1;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitCase(this, context);
    }

    override insideCaseStatement(): boolean {
        return true;
    }

    override indentedString(indent: string): string {
        const condition = (this.isDefaultCase) ? "default" : `case ${this.condition!.toString()}`;
        return `${indent}${condition}: ${this.statement.indentedStatementString(indent, true)}`;
    }



    // Backwards compatibility

    /**
     * @deprecated
     */
    getCondition(): CaseCondition | null {
        return this.condition;
    }

    /**
     * @deprecated
     */
    getExecution(): StatementWrapper {
        return this.statement;
    }
}