import { Node } from "./index";
import { Visitor } from "./../Visitor";
import { IllegalArgumentException } from "../../pseuco/constants";

export type ConditionModifier = "waitForCondition" | "signal" | "signalAll"

/**
 * AST node that represent the interaction with a condition object.
 * @example waitForCondition(c);
 * signal(c);
 * signalAll(c);
 * signalAll();
 */
export class ConditionStatement extends Node {
    readonly nodeName = "ConditionStatement";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param modifier The type of interaction with the condition, i.e., "waitForCondition", "signal" or "signalAll". 
     * @param identifier The name of the condition to interact with or *null* if signalAll shall interact with all conditions of the monitor.
     */
    constructor(line: number, column: number, readonly modifier: ConditionModifier, readonly identifier?: string) {
        super(line, column);
        if (identifier == undefined) {
            if (modifier != "signalAll") {
                throw IllegalArgumentException(`Expression may be omitted only for 'signalAll', not for '${modifier}'.`);
            }
        }
    }

    override indentedString(indent: string): string {
        if (this.identifier != null) {
            return `${indent}${this.modifier}(${this.identifier.toString()});`;
        } else {
            return `${indent}${this.modifier}();`;
        }
    }
    

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitConditionStatement(this, context);
    }
}