import { Node, StatementWrapper, ForInit, StatementExpression } from "./index";
import { Visitor } from "./../Visitor";
import { Expression } from "../typeNodes";

/**
 * AST node representing a `for` loop.
 * @example for (int x = 0; x < 42; x++) {
 *   println(x);
 * }
 */
export class ForStatement extends Node {
    readonly nodeName = "ForStatement";
    
    /**
     * The AST node representing the body of the loop.
     * @example {
     *   println(x);
     * }
     */
    readonly loopBody: StatementWrapper;

    /**
     * The AST node representing the initialisation of the loop.
     * @example int x = 0
     */
    readonly loopInit: ForInit | null;

    /**
     * The AST node representing the predicate expression that must evaluate to true to continue with another loop iteration.
     * @example x < 42
     */
    readonly predicate: Expression | null; 

    /**
     * An array of AST nodes that are executed after each loop iteration.
     * @example x++
     */
    readonly loopUpdates: StatementExpression[];

    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param loopBody The AST node representing the body of the loop.
     * @param loopInit The AST node representing the initialisation of the loop.
     * @param predicate The AST node representing the predicate expression that must evaluate to true to continue with another loop iteration.
     * @param loopUpdates An array of AST nodes that are executed after each loop iteration.
     */
    constructor(line: number, column: number, loopBody: StatementWrapper, loopInit: ForInit | null, predicate: Expression | null, ...loopUpdates: StatementExpression[]) {
        var children: Array<Node> = new Array<Node>();
        if (loopInit != null) {
            children.push(loopInit);
        }
        if (predicate != null) {
            children.push(predicate);
        }
        children = children.concat(loopUpdates);
        children.push(loopBody);
        super(line, column, ...children);
        this.loopBody = loopBody;
        this.loopInit = loopInit;
        this.predicate = predicate;
        this.loopUpdates = loopUpdates;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitForStatement(this, context);
    }

    override insideLoop(): boolean {
        return true;
    }

    override indentedString(indent: string): string {
        const initString = (this.loopInit != null) ? this.loopInit.toString() : "";
        const predicateString = (this.predicate != null) ? this.predicate.toString() : "";
        const updateString = this.loopUpdates.map(update => update.toString()).join(", ");
        return `${indent}for (${initString}; ${predicateString}; ${updateString}) ${this.loopBody.indentedStatementString(indent, true)}`;
    }
}