import { Node, Expression, StatementWrapper } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing an `if` statement.
 * @example if (predicate) {
 *   println("consequence");
 * } else {
 *   println("alternative");
 * }
 */
export class IfStatement extends Node {
    readonly nodeName;
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param predicate The AST node representing the predicate expression used to decide whether to continue with the consequence or (if specified) with the alternative.
     * @param consequence The AST node executed if the predicate evaluates to true.
     * @param alternative The AST node executed if the predicate evaluates to false. For *null*, the alternative is no operation. 
     */
    constructor(line: number, column: number, predicate: Expression, consequence: StatementWrapper, alternative?: StatementWrapper) {
        if (alternative == undefined) {
            super(line, column, predicate, consequence);
        } else {
            super(line, column, predicate, consequence, alternative);
        }
        this.nodeName = "IfStatement";
    }

    /**
     * The AST node representing the predicate expression used to decide whether to continue with the consequence or (if specified) with the alternative.
     */
    get predicate(): Expression {
        return this.childNodes[0] as Expression;
    }

    /**
     * The AST node executed if the predicate evaluates to true.
     */
    get consequence(): StatementWrapper {
        return this.childNodes[1] as StatementWrapper;
    }

    /**
     * The AST node executed if the predicate evaluates to false. For *null*, the alternative is no operation.
     */
    get alternative(): StatementWrapper | null {
        if (this.childNodes.length > 2) {
            return this.childNodes[2] as StatementWrapper;
        } else {
            return null;
        }
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitIfStatement(this, context);
    }

    override indentedString(indent: string): string {
        const elseCase = (this.alternative != null) ? `\n${indent}else ${this.alternative!.indentedStatementString(indent, true)}` : "";
        return `${indent}if (${this.predicate.toString()}) ${this.consequence.indentedStatementString(indent, true)} ${elseCase}`;
    }
}