import { Node } from "./index";
import { Visitor } from "./../Visitor";
import { Expression } from "../typeNodes";

export type Primitive = "join" | "lock" | "unlock" | "assert"

/**
 * AST node representing the primitive statements `join`, `lock`, `unlock` and `assert`.
 * @example join(agent);
 * lock(l);
 * unlock(l);
 * assert(X > 0);
 */
export class PrimitiveStatement extends Node {
    readonly nodeName = "PrimitiveStatement";
    
    private _primitive: Primitive

    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param primitive The concrete primitive that `this` node represents, i.e., "join", "lock", "unlock" or "assert".
     * @param expression The AST node representing the single argument that the primitive expects.
     */
    constructor(line: number, column: number, primitive: Primitive, expression: Expression) {
        super(line, column, expression);
        this._primitive = primitive;
    }

    /**
     * The concrete primitive that `this` node represents, i.e., "join", "lock", "unlock" or "assert".
     */
    get primitive(): Primitive {
        return this._primitive;
    }

    /**
     * The AST node representing the single argument that the primitive expects.
     */
    get expression(): Expression {
        return this.childNodes[0] as Expression;
    }

    override indentedString(indent: string): string {
        return `${indent}${this.primitive}(${this.expression.toString()});`;
    }
    

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitPrimitiveStatement(this, context);
    }
}