import { Node } from "./index";
import { Visitor } from "./../Visitor";
import { Expression } from "../expressions/Expression";

/**
 * AST node representing the `println` statement.
 * @example println("Hello World");  // prints "Hello World"
 * println("Hello ", "World");  // prints "Hello World"
 * println(4, 2);  // prints "42"
 */
export class PrintStatement extends Node {
    readonly nodeName = "PrintStatement";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param argumentExpressions One or more AST nodes representing expressions that evaluate to the arguments of this statement.
     */
    constructor(line: number, column: number, ...argumentExpressions: Expression[]) {
        super(line, column, ...argumentExpressions);
        if (argumentExpressions.length == 0) {
            throw new Error("println expects at least one argument.");
            // This constraint is already ensured by the parser and only necessary for manual construction of this node.
        }
    }

    /**
     * One or more AST nodes representing expressions that evaluate to the arguments of this statement. The arguments are string concatenated and printed as a single string.
     */
    get argumentExpressions(): Expression[] {
        return this.childNodes as Expression[];
    }

    override indentedString(indent: string): string {
        const argumentsString = this.argumentExpressions.map(function(expression) {
            return expression.toString();
        }).join(", ");
        return `${indent}println(${argumentsString});`;
    }
    

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitPrintStatement(this, context);
    }
}