import { Node } from "./index";
import { Visitor } from "./../Visitor";
import { Expression } from "../typeNodes";

/**
 * AST node representing the `return` statement.
 * @example return;
 * return 42;
 */
export class ReturnStatement extends Node {
    readonly nodeName;
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param expression If a value shall be returned, this argument is an AST node representing the expression that evaluates to the return value.
     */
    constructor(line: number, column: number, expression?: Expression) {
        if (expression) {
            super(line, column, expression);
        } else {
            super(line, column);
        }
        this.nodeName = "ReturnStatement";
    }

    /**
     * If a value shall be returned, this property is an AST node representing the expression that evaluates to the return value. Otherwise, the property is *null*.
     */
    get expression(): Expression | null {
        if (this.childNodes.length > 0) {
            return this.childNodes[0] as Expression;
        } else {
            return null;
        }
    }

    override indentedString(indent: string): string {
        const valString = (this.expression) ? ` ${this.expression.toString()}` : "";
        return `${indent}return${valString};`;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitReturnStatement(this, context);
    }
}