import { Node, Case } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing a `select` statement.
 * @example select {
 *   case <? chan: { }
 *   case x = <? chan: { }
 *   case chan <! val: { }
 *   default: { }
 * }
 */
export class SelectStatement extends Node {
    readonly nodeName = "SelectStatement";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param cases An array of AST nodes representing the cases of the select statement.
     */
    constructor(line: number, column: number, ...cases: Case[]) {
        super(line, column, ...cases);
        if (cases.length == 0) {
            throw new Error("A select statement must have at least one case.");
            // This constraint is already ensured by the parser and only necessary for manual construction of this node.
        }
    }

    /**
     * An array of AST nodes representing the cases of the select statement.
     */
    get cases(): Case[] {
        return this.childNodes as Case[];
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitSelectStatement(this, context);
    }

    override indentedString(indent: string): string {
        const childrenString = this.cases.map(function (c: Case) {
            return c.indentedString(indent + Node.INDENT);
        }).join("\n");
        return `${indent}select {\n${childrenString}\n${indent}}`;
    }
}

