import { Node, StatementWrapper, VariableDeclarations, ConditionDeclaration } from "./index";
import { Visitor } from "./../Visitor";

export type BlockStatement = StatementWrapper | VariableDeclarations | ConditionDeclaration;

/**
 * AST node representing a set of statements wrapped in a pair of curly braces.
 * @example {
 *   // none, one or many statements
 * }
 */
export class StatementBlock extends Node {
    readonly nodeName = "StatementBlock";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param children An array of AST nodes that represent a statement (see [StatementWrapper](./StatementWrapper.ts) for details), variable or condition declarations. 
     */
    constructor(line: number, column: number, ...children: BlockStatement[]) {
        super(line, column, ...children);
    }

    /**
     * An array of AST nodes that represent a statement (see [StatementWrapper](./StatementWrapper.ts) for details), variable or condition declarations. 
     */
    get blockStatements(): BlockStatement[] {
        return this.childNodes as BlockStatement[];
    }

    override indentedString(indent: string): string {
        const childrenString = this.blockStatements.map(function(statementOrDeclaration) {
            return statementOrDeclaration.indentedString(indent + Node.INDENT);
        }).join("\n");
        return `${indent}{\n${childrenString}\n${indent}}`;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitStatementBlock(this, context);
    }
}