import { Node, Expression, StartExpression, AssignExpression, SendExpression, PostfixExpression, ProcedureCall, ClassCall, ReceiveExpression } from "./index";
import { Visitor } from "./../Visitor";

export type StatementableExpression = StartExpression | AssignExpression | SendExpression | PostfixExpression | ProcedureCall | ClassCall | ReceiveExpression;

/**
 * AST node that represents expressions that can be turned into statements by adding a semicolon behind it.
 * @example start(f())
 * x = 5
 * chan <! 42
 * <? chan
 * i++
 * f()
 * m.f()
 */
export class StatementExpression extends Node {
    readonly nodeName = "StatementExpression";
    
    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param expression AST node representing a start, assign, send, receive, postfix or procedure call expression.
     */
    constructor(line: number, column: number, expression: StatementableExpression) {
        super(line, column, expression);
    }

    /**
     * AST node representing a start, assign, send, receive, postfix or procedure call expression.
     */
    get expression(): StatementableExpression {
        return this.childNodes[0] as StatementableExpression;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitStatementExpression(this, context);
    }

    override indentedString(indent: string): string {
        return `${indent}${this.expression.toString()}`;
    }
}