import { Node, VariableDeclarations, StatementBlock, BreakStatement, ContinueStatement, DoStatement, ForStatement, IfStatement, PrimitiveStatement, PrintStatement, ReturnStatement, SelectStatement, StatementExpression, WhileStatement, ConditionStatement } from "./index";
import { Visitor } from "./../Visitor";

export type Statement = StatementBlock | BreakStatement | ContinueStatement | DoStatement | ForStatement | IfStatement | PrimitiveStatement | ConditionStatement | PrintStatement | ReturnStatement | SelectStatement | StatementExpression | WhileStatement;
// Not wrappable are objects of classes Case, ForInit, and StatementWrapper

/**
 * An intermediate AST node to properly manage statement nodes. Allows the empty (noop) statement. Adds the semicolon behind *StatementExpression*s.
 * @example ;  // noop
 * x = 5;  // statement expression
 * println(42);
 * // ...
 */
export class StatementWrapper extends Node {
    readonly nodeName: string;
    
    /**
     * 
    * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param statement For non-noop statements, the statement that shall be wrapped. Can be a *StatementBlock*, *BreakStatement*, *ContinueStatement*, *DoStatement*, *ForStatement*, *IfStatement*, *PrimitiveStatement*, *ConditionStatement*, *PrintStatement*, *ReturnStatement*, *SelectStatement*, *StatementExpression* or *WhileStatement*.
     */
    constructor(line: number, column: number, statement?: Statement) {
        if (statement) {
            super(line, column, statement);
        } else {
            super(line, column);
        }
        this.nodeName = "StatementWrapper";
    }

    /**
     * If the wrapped statement is a statement block, holds the statement block. Otherwise it is *null*.
     */
    get statementBlock(): StatementBlock | null {
        if (this.childNodes.length > 0 && this.childNodes[0] instanceof StatementBlock) {
            return this.childNodes[0] as StatementBlock;
        } else {
            return null;
        }
    }

    /**
     * If the wrapped statement is a statement expression, holds the statement expression. Otherwise it is *null*.
     */
    get statementExpression(): StatementExpression | null {
        if (this.childNodes.length > 0 && this.childNodes[0] instanceof StatementExpression) {
            return this.childNodes[0] as StatementExpression;
        } else {
            return null;
        }
    }

    /**
     * *null* for noop statements and otherwise the statement that shall be wrapped. Can be a *StatementBlock*, *BreakStatement*, *ContinueStatement*, *DoStatement*, *ForStatement*, *IfStatement*, *PrimitiveStatement*, *ConditionStatement*, *PrintStatement*, *ReturnStatement*, *SelectStatement*, *StatementExpression* or *WhileStatement*.
     */
    get statement(): Statement | null {
        if (this.childNodes.length > 0) {
            return this.childNodes[0] as Statement;
        } else {
            return null;
        }
    }

    /**
     * Flag indicating whether this node represents a noop (i.e., an empty) statement.
     */
    get isNoOp(): boolean {
        return this.childNodes.length == 0;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitStatementWrapper(this, context);
    }

    /**
     * Computes a string of the wrapped statements. This method takes special care about the different handling of single line statements and statement blocks wrapped in curly braces. The former sometimes want a newline and extra indentation level before their string representation, the latter don't.
     * @param indent The indentation level that shall be used at the beginning of each line.
     * @param increaseIndentationLevelForChildren Flag whether to increase the indentation level for single line statements (i.e., for nodes that are not a *StatementBlock*)
     * @returns A correctly indented string representation of the wrapped statement.
     * @example if (predicate)
     *   println("This is a single line statement that got a newline and extra indent.");
     * 
     * if (predicate) {
     *   println("This is a three-line statement block that got no newline and no extra indent.");
     * }
     */
    indentedStatementString(indent: string, increaseIndentationLevelForChildren: boolean): string {
        let newLine = "";
        if (increaseIndentationLevelForChildren && (this.childNodes.length == 0 || !(this.childNodes[0] instanceof StatementBlock))) {
            indent += Node.INDENT;
            newLine = "\n";
        }

        if (this.isNoOp) {
            return `${newLine}${indent};`;
        } else {
            const childString = this.statement!.indentedString(indent);
            const semicolon = (this.statementExpression != null) ? ";" : "";
            return `${newLine}${childString}${semicolon}`;
        }
    }

    override indentedString(indent: string): string {
        return this.indentedStatementString(indent, false);
    }
}
