import { Node, Expression, StatementWrapper } from "./index";
import { Visitor } from "./../Visitor";

/**
 * AST node representing a while loop.
 * @example while (predicate) {
 *   // statements
 * }
 */
export class WhileStatement extends Node {
    readonly nodeName = "WhileStatement";
    

    /**
     * 
     * @param line The (first) line number in which `this` node appears in the pseuCo program.
     * @param column The first column number at which `this` node appears in the pseuCo program.
     * @param predicate The AST node representing the predicate expression that must evaluate to true to continue with another loop iteration.
     * @param loopBody The AST node representing the loop body.
     */
    constructor(line: number, column: number, predicate: Expression, loopBody: StatementWrapper) {
        super(line, column, predicate, loopBody);
    }

    /**
     * The AST node representing the predicate expression that must evaluate to true to continue with another loop iteration.
     */
    get predicate(): Expression {
        return this.childNodes[0] as Expression;
    }

    /**
     * The AST node representing the loop body.
     */
    get loopBody(): StatementWrapper {
        return this.childNodes[1] as StatementWrapper;
    }

    override accept<T>(visitor: Visitor<T>, context?: unknown): T {
        return visitor.visitWhileStatement(this, context);
    }

    override insideLoop(): boolean {
        return true;
    }

    override indentedString(indent: string): string {
        return `${indent}while (${this.predicate.toString()}) ${this.loopBody.indentedStatementString(indent, true)}`;
    }
}